@@ -70,7 +70,7 @@ You shall use the default value if using auto-hide minibuffer."
 
 (defvar exwm-systemtray--connection nil "The X connection.")
 
-(defvar exwm-systemtray--embedder nil "The embedder window.")
+(defvar exwm-systemtray--embedder-window nil "The embedder window.")
 
 (defvar exwm-systemtray--list nil "The icon list.")
 
@@ -86,13 +86,13 @@ You shall use the default value if using auto-hide minibuffer."
 
 (defun exwm-systemtray--embed (icon)
   "Embed an icon."
-  (exwm--log "(System Tray) Try to embed #x%x" icon)
+  (exwm--log "Try to embed #x%x" icon)
   (let ((info (xcb:+request-unchecked+reply exwm-systemtray--connection
                   (make-instance 'xcb:xembed:get-_XEMBED_INFO
                                  :window icon)))
         width* height* visible)
     (when info
-      (exwm--log "(System Tray) Embed #x%x" icon)
+      (exwm--log "Embed #x%x" icon)
       (with-slots (width height)
           (xcb:+request-unchecked+reply exwm-systemtray--connection
               (make-instance 'xcb:GetGeometry :drawable icon))
@@ -101,7 +101,7 @@ You shall use the default value if using auto-hide minibuffer."
         (when (< width* exwm-systemtray--icon-min-size)
           (setq width* exwm-systemtray--icon-min-size
                 height* (round (* height (/ (float width*) width)))))
-        (exwm--log "(System Tray) Resize from %dx%d to %dx%d"
+        (exwm--log "Resize from %dx%d to %dx%d"
                    width height width* height*))
       ;; Add this icon to save-set.
       (xcb:+request exwm-systemtray--connection
@@ -112,7 +112,7 @@ You shall use the default value if using auto-hide minibuffer."
       (xcb:+request exwm-systemtray--connection
           (make-instance 'xcb:ReparentWindow
                          :window icon
-                         :parent exwm-systemtray--embedder
+                         :parent exwm-systemtray--embedder-window
                          :x 0
                          ;; Vertically centered.
                          :y (/ (- exwm-systemtray-height height*) 2)))
@@ -151,7 +151,7 @@ You shall use the default value if using auto-hide minibuffer."
         ;; Default to visible.
         (setq visible t))
       (when visible
-        (exwm--log "(System Tray) Map the window")
+        (exwm--log "Map the window")
         (xcb:+request exwm-systemtray--connection
             (make-instance 'xcb:MapWindow :window icon)))
       (xcb:+request exwm-systemtray--connection
@@ -162,7 +162,8 @@ You shall use the default value if using auto-hide minibuffer."
                           (make-instance 'xcb:xembed:EMBEDDED-NOTIFY
                                          :window icon
                                          :time xcb:Time:CurrentTime
-                                         :embedder exwm-systemtray--embedder
+                                         :embedder
+                                         exwm-systemtray--embedder-window
                                          :version 0)
                           exwm-systemtray--connection)))
       (push `(,icon . ,(make-instance 'exwm-systemtray--icon
@@ -174,7 +175,7 @@ You shall use the default value if using auto-hide minibuffer."
 
 (defun exwm-systemtray--unembed (icon)
   "Unembed an icon."
-  (exwm--log "(System Tray) Unembed #x%x" icon)
+  (exwm--log "Unembed #x%x" icon)
   (xcb:+request exwm-systemtray--connection
       (make-instance 'xcb:UnmapWindow :window icon))
   (xcb:+request exwm-systemtray--connection
@@ -188,9 +189,11 @@ You shall use the default value if using auto-hide minibuffer."
 
 (defun exwm-systemtray--refresh ()
   "Refresh the system tray."
+  (exwm--log)
   ;; Make sure to redraw the embedder.
   (xcb:+request exwm-systemtray--connection
-      (make-instance 'xcb:UnmapWindow :window exwm-systemtray--embedder))
+      (make-instance 'xcb:UnmapWindow
+                     :window exwm-systemtray--embedder-window))
   (let ((x exwm-systemtray-icon-gap)
         map)
     (dolist (pair exwm-systemtray--list)
@@ -207,18 +210,20 @@ You shall use the default value if using auto-hide minibuffer."
                          exwm-workspace-current-index)))
       (xcb:+request exwm-systemtray--connection
           (make-instance 'xcb:ConfigureWindow
-                         :window exwm-systemtray--embedder
+                         :window exwm-systemtray--embedder-window
                          :value-mask (logior xcb:ConfigWindow:X
                                              xcb:ConfigWindow:Width)
                          :x (- (aref workarea 2) x)
                          :width x)))
     (when map
       (xcb:+request exwm-systemtray--connection
-          (make-instance 'xcb:MapWindow :window exwm-systemtray--embedder))))
+          (make-instance 'xcb:MapWindow
+                         :window exwm-systemtray--embedder-window))))
   (xcb:flush exwm-systemtray--connection))
 
 (defun exwm-systemtray--on-DestroyNotify (data _synthetic)
   "Unembed icons on DestroyNotify."
+  (exwm--log)
   (let ((obj (make-instance 'xcb:DestroyNotify)))
     (xcb:unmarshal obj data)
     (with-slots (window) obj
@@ -227,15 +232,17 @@ You shall use the default value if using auto-hide minibuffer."
 
 (defun exwm-systemtray--on-ReparentNotify (data _synthetic)
   "Unembed icons on ReparentNotify."
+  (exwm--log)
   (let ((obj (make-instance 'xcb:ReparentNotify)))
     (xcb:unmarshal obj data)
     (with-slots (window parent) obj
-      (when (and (/= parent exwm-systemtray--embedder)
+      (when (and (/= parent exwm-systemtray--embedder-window)
                  (assoc window exwm-systemtray--list))
         (exwm-systemtray--unembed window)))))
 
 (defun exwm-systemtray--on-ResizeRequest (data _synthetic)
   "Resize the tray icon on ResizeRequest."
+  (exwm--log)
   (let ((obj (make-instance 'xcb:ResizeRequest))
         attr)
     (xcb:unmarshal obj data)
@@ -263,6 +270,7 @@ You shall use the default value if using auto-hide minibuffer."
 
 (defun exwm-systemtray--on-PropertyNotify (data _synthetic)
   "Map/Unmap the tray icon on PropertyNotify."
+  (exwm--log)
   (let ((obj (make-instance 'xcb:PropertyNotify))
         attr info visible)
     (xcb:unmarshal obj data)
@@ -276,7 +284,7 @@ You shall use the default value if using auto-hide minibuffer."
         (when info
           (setq visible (/= 0 (logand (slot-value info 'flags)
                                       xcb:xembed:MAPPED)))
-          (exwm--log "(System Tray) #x%x visible? %s" window visible)
+          (exwm--log "#x%x visible? %s" window visible)
           (if visible
               (xcb:+request exwm-systemtray--connection
                   (make-instance 'xcb:MapWindow :window window))
@@ -294,6 +302,7 @@ You shall use the default value if using auto-hide minibuffer."
       (when (eq type xcb:Atom:_NET_SYSTEM_TRAY_OPCODE)
         (setq data32 (slot-value data 'data32)
               opcode (elt data32 1))
+        (exwm--log "opcode: %s" opcode)
         (cond ((= opcode xcb:systemtray:opcode:REQUEST-DOCK)
                (unless (assoc (elt data32 2) exwm-systemtray--list)
                  (exwm-systemtray--embed (elt data32 2))))
@@ -301,10 +310,11 @@ You shall use the default value if using auto-hide minibuffer."
               ((or (= opcode xcb:systemtray:opcode:BEGIN-MESSAGE)
                    (= opcode xcb:systemtray:opcode:CANCEL-MESSAGE)))
               (t
-               (exwm--log "(System Tray) Unknown opcode message: %s" obj)))))))
+               (exwm--log "Unknown opcode message: %s" obj)))))))
 
 (defun exwm-systemtray--on-KeyPress (data _synthetic)
   "Forward all KeyPress events to Emacs frame."
+  (exwm--log)
   ;; This function is only executed when there's no autohide minibuffer,
   ;; a workspace frame has the input focus and the pointer is over a
   ;; tray icon.
@@ -322,10 +332,11 @@ You shall use the default value if using auto-hide minibuffer."
 
 (defun exwm-systemtray--on-workspace-switch ()
   "Reparent/Refresh the system tray in `exwm-workspace-switch-hook'."
+  (exwm--log)
   (unless (exwm-workspace--minibuffer-own-frame-p)
     (xcb:+request exwm-systemtray--connection
         (make-instance 'xcb:ReparentWindow
-                       :window exwm-systemtray--embedder
+                       :window exwm-systemtray--embedder-window
                        :parent (string-to-number
                                 (frame-parameter exwm-workspace--current
                                                  'window-id))
@@ -336,10 +347,11 @@ You shall use the default value if using auto-hide minibuffer."
 
 (defun exwm-systemtray--on-randr-refresh ()
   "Reposition/Refresh the system tray in `exwm-randr-refresh-hook'."
+  (exwm--log)
   (unless (exwm-workspace--minibuffer-own-frame-p)
     (xcb:+request exwm-systemtray--connection
         (make-instance 'xcb:ConfigureWindow
-                       :window exwm-systemtray--embedder
+                       :window exwm-systemtray--embedder-window
                        :value-mask xcb:ConfigWindow:Y
                        :y (- (frame-pixel-height exwm-workspace--current)
                              exwm-systemtray-height))))
@@ -350,10 +362,11 @@ You shall use the default value if using auto-hide minibuffer."
 
 (cl-defun exwm-systemtray--init ()
   "Initialize system tray module."
+  (exwm--log)
   (cl-assert (not exwm-systemtray--connection))
   (cl-assert (not exwm-systemtray--list))
   (cl-assert (not exwm-systemtray--selection-owner-window))
-  (cl-assert (not exwm-systemtray--embedder))
+  (cl-assert (not exwm-systemtray--embedder-window))
   (unless exwm-systemtray-height
     (setq exwm-systemtray-height (max exwm-systemtray--icon-min-size
                                       (line-pixel-height))))
@@ -414,7 +427,8 @@ You shall use the default value if using auto-hide minibuffer."
     ;; Set _NET_WM_NAME.
     (xcb:+request exwm-systemtray--connection
         (make-instance 'xcb:ewmh:set-_NET_WM_NAME
-                       :window id :data "EXWM system tray selection owner"))
+                       :window id
+                       :data "EXWM: exwm-systemtray--selection-owner-window"))
     ;; Set the _NET_SYSTEM_TRAY_ORIENTATION property.
     (xcb:+request exwm-systemtray--connection
         (make-instance 'xcb:xembed:set-_NET_SYSTEM_TRAY_ORIENTATION
@@ -423,7 +437,7 @@ You shall use the default value if using auto-hide minibuffer."
   ;; Create the embedder.
   (let ((id (xcb:generate-id exwm-systemtray--connection))
         frame parent depth y)
-    (setq exwm-systemtray--embedder id)
+    (setq exwm-systemtray--embedder-window id)
     (if (exwm-workspace--minibuffer-own-frame-p)
         (setq frame exwm-workspace--minibuffer
               y (if (>= (line-pixel-height) exwm-systemtray-height)
@@ -460,7 +474,8 @@ You shall use the default value if using auto-hide minibuffer."
     ;; Set _NET_WM_NAME.
     (xcb:+request exwm-systemtray--connection
         (make-instance 'xcb:ewmh:set-_NET_WM_NAME
-                       :window id :data "EXWM system tray embedder")))
+                       :window id
+                       :data "EXWM: exwm-systemtray--embedder-window")))
   (xcb:flush exwm-systemtray--connection)
   ;; Attach event listeners.
   (xcb:+event exwm-systemtray--connection 'xcb:DestroyNotify
@@ -488,16 +503,17 @@ You shall use the default value if using auto-hide minibuffer."
 
 (defun exwm-systemtray--exit ()
   "Exit the systemtray module."
+  (exwm--log)
   (when exwm-systemtray--connection
     ;; Hide & reparent out the embedder before disconnection to prevent
     ;; embedded icons from being reparented to an Emacs frame (which is the
     ;; parent of the embedder).
     (xcb:+request exwm-systemtray--connection
         (make-instance 'xcb:UnmapWindow
-                       :window exwm-systemtray--embedder))
+                       :window exwm-systemtray--embedder-window))
     (xcb:+request exwm-systemtray--connection
         (make-instance 'xcb:ReparentWindow
-                       :window exwm-systemtray--embedder
+                       :window exwm-systemtray--embedder-window
                        :parent exwm--root
                        :x 0
                        :y 0))
@@ -505,7 +521,7 @@ You shall use the default value if using auto-hide minibuffer."
     (setq exwm-systemtray--connection nil
           exwm-systemtray--list nil
           exwm-systemtray--selection-owner-window nil
-          exwm-systemtray--embedder nil)
+          exwm-systemtray--embedder-window nil)
     (remove-hook 'exwm-workspace-switch-hook
                  #'exwm-systemtray--on-workspace-switch)
     (remove-hook 'exwm-workspace--update-workareas-hook
@@ -516,6 +532,7 @@ You shall use the default value if using auto-hide minibuffer."
 
 (defun exwm-systemtray-enable ()
   "Enable system tray support for EXWM."
+  (exwm--log)
   (add-hook 'exwm-init-hook #'exwm-systemtray--init)
   (add-hook 'exwm-exit-hook #'exwm-systemtray--exit))
 
